import std.c.stdio;

// 2 sorting algs
interface SortAlg {
	void sort(int n);
	int  cmp(int i,int j);
	void swp(int i,int j);
};
abstract class PiramidSort : SortAlg {
	void sort(int n) { // O(n*ln(n))  
	  for(int i=n/2;i>=0;i--) rise(i,n); 
	  for(int i=n-1;i>=0;i--) { swp(0,i); rise(0,i); }
	}
	final void rise(int i,int n) {
		int c; for(int r=i;r*2<n;r=c) {
	      c=r*2; if (c<n-1 && cmp(c,c+1)<0 ) c++;
	      if (cmp(r,c)>=0) break; swp(r,c);  
	    }
	}
}
abstract class QuickSort : SortAlg {
	void sort(int n) { // from C stdlib ~O(n*ln(n))
		int[16*int.sizeof-6] stack;
		int *sp=&stack[0]; int i,j,L=0,R=n-1,thresh=7;
		for(;;) {
			while (R-L>=thresh) {
				swp((L+R)>>1,L); i=L+1; j=R;
				if (cmp(i,j)>0) swp(i,j);
				if (cmp(L,j)>0) swp(L,j);
				if (cmp(i,L)>0) swp(i,L);
				for(;;) {
					do i++; while (cmp(i,L)<0); do j--; while (cmp(j,L)>0);
					if (i>j) break; else swp(i,j);
				}
				swp(L,j);
				if (j-L > R-i) { sp[0]=L; sp[1]=j; L=i; } else { sp[0]=i; sp[1]=R; R=j; }
				sp+=2;
			}
			i=L+1;
			while (i<=R) { j=i; while (j>L && cmp(j-1,j)>0 ) { swp(j-1,j); j--; } i++; }
			if (sp>&stack[0]) { sp-=2; L=sp[0]; R=sp[1]; } else break;
		}
	}
}

// 6 victims types
// cmp.v2
struct MyType1 {
	int a=0,b=0;
	int opCmp(ref MyType1 rv) { // build in sort fails
		if (a!=rv.a) return a<rv.a?-1:+1;
		if (b!=rv.b) return b<rv.b?-1:+1;
		return 0;		
	}
};
struct MyType2 {
	int a=0,b=0;
	int opCmp(MyType2 rv) { // build in sort fails
		if (a!=rv.a) return a<rv.a?-1:+1;
		if (b!=rv.b) return b<rv.b?-1:+1;
		return 0;		
	}
};
struct MyType3 {
	int a=0,b=0;
	int opCmp(MyType3 *rv) { // can not find opCmp
		if (a!=rv.a) return a<rv.a?-1:+1;
		if (b!=rv.b) return b<rv.b?-1:+1;
		return 0;		
	}
};
// cmp.v1 with substraction
struct MyType4 { 
	int a=0,b=0;
	int opCmp(ref MyType4 rv) {
		if (a!=rv.a) return a-rv.a;
		if (b!=rv.b) return b-rv.b;
		return 0;		
	}
};
struct MyType5 { 
	int a=0,b=0;
	int opCmp(MyType5 rv) { 
		if (a!=rv.a) return a-rv.a;
		if (b!=rv.b) return b-rv.b;
		return 0;		
	}
};
struct MyType6 { // can not find opCmp 
	int a=0,b=0;
	int opCmp(MyType6 *rv) { // can not find opCmp
		if (a!=rv.a) return a-rv.a;
		if (b!=rv.b) return b-rv.b;
		return 0;		
	}
};

// try to use templates, but without result. So simply change params here by hands
alias MyType2     Type;           // select victim and sort alg
alias PiramidSort DefaultSortAlg;
//alias QuickSort DefaultSortAlg;

class DirectSort : DefaultSortAlg {
	Type[] data;
	this(Type[] x) { data=x; sort(x.length); }
	void swp(int i,int j) { auto t=data[i];data[i]=data[j];data[j]=t; }
	int  cmp(int i,int j) { return -data[i].opCmp(data[j]);            }
};
class IndirectSort : DefaultSortAlg {
	Type[] data; int[] idx;
	this(Type[] x) { int n=x.length; data=x; idx=new int[n]; for(int i=0;i<n;i++) idx[i]=i; sort(n); }
	void swp(int i,int j) { auto t=idx[i];idx[i]=idx[j];idx[j]=t;    }
	int  cmp(int i,int j) { return -data[idx[i]].opCmp(data[idx[j]]); }
};

int chk_less(Type a,Type b) {
	if (a.a< b.a) return true;
	if (a.a==b.a) return a.b<b.b;
	return false;
}
bool show(Type[] x) { 
	bool valid=true;
	for(int i=0;i<x.length;i++) {
		if (i>0 && chk_less(x[i-1],x[i])) valid=false;
		printf("%d. %11d.%11d\n",i,x[i].a,x[i].b); 
	}
	return valid;
}
bool ishow(Type[] x,int[] idx) { 
	bool valid=true;
	for(int i=0;i<idx.length;i++) {
		if (i>0 && chk_less(x[idx[i-1]],x[idx[i]])) valid=false;
		printf("%d. %11d.%11d\n",i,x[idx[i]].a,x[idx[i]].b); 
	}
	return valid;
}

int main(char[][] args) {
	bool[3][2] res;
	Type[] x=new Type[7];
	x[0].a=0;x[0].b=-0x8000_0000;
	x[1].a=0;x[1].b=-0x5000_0000;
	x[2].a=0;x[2].b=-0x3000_0000;
	x[3].a=0;x[3].b= 0x0000_0000;
	x[4].a=0;x[4].b=+0x3000_0000;
	x[5].a=0;x[5].b=+0x4FFF_FFFF;
	x[6].a=0;x[6].b=+0x7FFF_FFFF;
	// verbose
	for(int j=0;j<3;j++) {
		printf("indirect sort\n");
		IndirectSort ids=new IndirectSort(x);
		res[j][0]=ishow(x,ids.idx);
		
		printf("direct sort\n");
		DirectSort ds=new DirectSort(x);
		res[j][1]=show(x);
		
		printf("buildin sort\n");
		x.sort;
		res[j][2]=show(x);
	
		for(int i=0;i<x.length;i++) x[i].b=(i^5)+((j==1)?0:-3); // simple case
	}
	// brief info
	char[][4] hdr=["general","simple+","simple",""];
	printf("%7s | D I B\n",&hdr[3][0]);
	for(int j=0;j<3;j++) {
		printf("%7s |",&hdr[j][0]);
		for(int i=0;i<3;i++) {
			printf((res[j][i])?" +":" -");
		}
		printf("\n");
	}	
	return +1;
}

/*

cmp.v2

	MyType1 | D I B
	general | + + -
	 simple | + + +
	 
	MyType2 | D I B
	general | + + -
	simple+ | + + -
	 simple | + + +

	MyType3 -- compile error
 
cmp.v1

	MyType4 | D I B
	general | - - -
	simple+ | + + -
	 simple | + + +
	
	MyType5 | D I B
	general | - - -
	simple+ | + + -
	 simple | + + +

	MyType6 -- compile error
 
D - DirectSort
I - IndirectSort
B - buildin sort
 
*/
